from __future__ import print_function

from googleapiclient import discovery
from httplib2 import Http
from oauth2client import file, client
from oauth2client.tools import argparser, run_flow
import argparse
import os

SCOPES = 'https://www.googleapis.com/auth/drive.readonly.metadata'
store = file.Storage(os.path.expanduser("~") + '/.gdrive_credential.json')
creds = store.get()
if not creds or creds.invalid:
    flow = client.flow_from_clientsecrets(os.path.expanduser("~") + '/.rtutils_credentials.json', SCOPES)
    args = argparser.parse_args()
    args.noauth_local_webserver = True
    creds = run_flow(flow, store, args)
DRIVE = discovery.build('drive', 'v3', http=creds.authorize(Http()))

files = DRIVE.files().list().execute().get('files', [])
for f in files:
    print(f['name'], f['mimeType'])


import os
import mmcv

def get_coco_filename(cocoid):
    cocoid = int(cocoid)
    if os.path.isfile('/share/data/vdata/coco/images/train2017/%012d.jpg' %(cocoid)):
        return '/share/data/vdata/coco/images/train2017/%012d.jpg' %(cocoid)
    else:
        return '/share/data/vdata/coco/images/val2017/%012d.jpg' %(cocoid)

coco_image_infos = None

def get_coco_image_infos():
    coco_image_infos = mmcv.load('/share/data/vdata/coco/annotations/captions_train2017.json')['images'] + mmcv.load('/share/data/vdata/coco/annotations/captions_val2017.json')['images']
    coco_image_infos = {_['id']: _ for _ in coco_image_infos}
    return coco_image_infos


def get_coco_url(cocoid):
    global coco_image_infos
    if coco_image_infos is None:
        coco_image_infos = get_coco_image_infos()
    cocoid = int(cocoid)
    return coco_image_infos[cocoid]['coco_url']


import argparse

from pydrive.auth import GoogleAuth
from pydrive.drive import GoogleDrive
from oauth2client.client import GoogleCredentials
import os


class MyDrive:
    def __init__(self):
        gauth = GoogleAuth()
        cred_file = os.path.expanduser("~") + '/.gdrive_credential.json'
        if os.path.exists(cred_file):
            gauth.LoadCredentialsFile(cred_file)
        else:
            gauth.LoadClientConfigFile(os.path.expanduser("~") + '/.rtutils_credentials.json')
            gauth.CommandLineAuth()
            # gauth.credentials = '~/.rtutils_client.json' #GoogleCredentials.get_application_default()
            gauth.SaveCredentialsFile(cred_file)

        self.drive = GoogleDrive(gauth)

    def create_folder(self, folder, remote_folder_id):
        newFolder = self.drive.CreateFile({'title': folder, "parents": [{"kind": "drive#fileLink", "id": remote_folder_id}],
                                    "mimeType": "application/vnd.google-apps.folder"})
        newFolder.Upload()
        return newFolder['id']

    def upload_folder(self, root, folder, remote_folder_id):
        # Create folder
        newFolder_id = self.create_folder(folder, remote_folder_id)
        new_root = os.path.join(root, folder)
        for item in os.listdir(new_root):
            if os.path.isfile(os.path.join(new_root, item)):
                # Upload files in this folder
                self.upload_file(new_root, item, newFolder_id)
            else:
                # Upload folder in this folder
                self.upload_folder(new_root, item, newFolder_id)
        return newFolder_id

    def upload_file(self, root, filename, remote_folder_id=None):
        if remote_folder_id == None:
            newFile = self.drive.CreateFile({"title": filename})
        else:
            newFile = self.drive.CreateFile({"title": filename,
                                             "parents": [{"kind": "drive#fileLink", "id": remote_folder_id}]})
        newFile.SetContentFile(os.path.join(root, filename))
        newFile.Upload()

    def download_file(self, download_id, download_dir):
        newFile = self.drive.CreateFile({'id': download_id})
        newFile.FetchMetadata(fetch_all=True)
        newFile.GetContentFile(os.path.join(download_dir, newFile['title']))
        return newFile

def main():
    drive = MyDrive()
    parser = argparse.ArgumentParser(description='Drive')
    subparsers = parser.add_subparsers(dest="command", required=True)
    parser_upload = subparsers.add_parser('upload', help='upload file or folder')
    parser_download = subparsers.add_parser('download', help='download file')

    parser_upload.add_argument('upload_path', type=str, help='')
    parser_upload.add_argument('remote_folder_id', type=str, nargs='?', default=None, help='')

    parser_download.add_argument('download_id', type=str, help='')
    parser_download.add_argument('download_dir', type=str, nargs='?', default='./', help='')

    args = parser.parse_args()
    # print(args)

    if args.command == 'upload':
        path = args.upload_path
        root = os.path.dirname(path)
        name = os.path.basename(path)
        if os.path.isdir(path):
            drive.upload_folder(root, name, args.remote_folder_id)
        else:
            drive.upload_file(root, name, args.remote_folder_id)
        print('Upload done')
    elif args.command == 'download':
        file = drive.download_file(args.download_id, args.download_dir)
        print('Download finished %s' %os.path.join(args.download_dir, file["title"]))

#     file_name = args.file
#     action = args.action
#     dummy = args.dummy
#     partition = args.partition
#     length = args.length
#     num_cores = args.num_cores
#     features = args.feature_constraints
#     print("Using partition {}".format(partition))
#     if dummy:
#         print("Under dummy mode")
#     if action not in allowed_actions:
#         raise ValueError(
#             "action must be one of {}, but given: {}".format(allowed_actions, action))

#     with open(file_name) as f:
#         task_dir_list = yaml.load(f)
#     for task_dir in task_dir_list:
#         if not path.isdir(task_dir):
#             raise ValueError("{} is not a valid directory".format(task_dir))
#         else:
#             task_execute(
#                 task_dir, action, length, dummy, partition, num_cores, features)

if __name__ == '__main__':
    drive = MyDrive()
    main()
    # drive.upload_folder('../content/', 'weekly_2020-01-10', '1DagXOiUK-oqBQ7lN734X1ZdxAiRnsFC1')


import torch
import itertools
import time
import torch.distributed as dist

class ResumableDataLoader(torch.utils.data.DataLoader):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._need_to_resume = False
        self._rng_state = None
        self._iter = None

    def get_sampler_iter(self):
        tmp_num_workers = self.num_workers
        self.num_workers = 0
        dataiter = super().__iter__()
        __sampler_iter, _sampler_iter = itertools.tee(dataiter._sampler_iter)
        next(__sampler_iter) # Make sure the _sampler_iter is instantiated
        self.num_workers = tmp_num_workers
        return _sampler_iter

    def __iter__(self):

        if self._need_to_resume:
            # Run through the saved_iter, make the sampler at the state of halt
            self._sampler_iter, _sampler_iter = itertools.tee(iter(self._saved_sampler))
            for i in range(self._saved_iter):
                next(_sampler_iter)
        else:
            _sampler_iter = self.get_sampler_iter()
            self._sampler_iter, _sampler_iter = itertools.tee(_sampler_iter) # Save the full iter

        self.get_sampler_to_save() # for state_dict

        dataiter = super().__iter__()
        dataiter._sampler_iter = _sampler_iter
        if self.num_workers > 0:
            # Wait for prefetching
            time.sleep(5)
            # drain the prefetched batches
            for i in range(dataiter._send_idx - dataiter._rcvd_idx):
                next(dataiter)

        self._iter = dataiter # Save it in case
        self._need_to_resume = False # Only resume at the first iter(loader) after load_state_dict
        return self._iter

    def get_sampler_to_save(self):
        # All the indices:
        self._sampler_to_save = []
        for idxs in self._sampler_iter:
            if dist.is_available() and type(self.sampler) is torch.utils.data.distributed.DistributedSampler:
                idxs_list = [torch.zeros(len(idxs)).long().to(torch.cuda.current_device()) for k in range(dist.get_world_size())]
                dist.all_gather(idxs_list, torch.tensor(idxs).long().to(torch.cuda.current_device()), async_op=False) # can also use detectron.utils.comm.all_gather
                self._sampler_to_save.append([_.tolist() for _ in idxs_list])
            else:
                self._sampler_to_save.append(idxs)

    def state_dict(self):
        if self._iter is None:
            return None
        # The number of batches rest in this epoch
        num_rest = 0
        self._iter._sampler_iter, _sampler_iter = itertools.tee(self._iter._sampler_iter)
        for idxs in _sampler_iter:
            num_rest += 1

        # the number of batched prefetched by the loader
        if self.num_workers > 0:
            num_prefetched = self._iter._send_idx - self._iter._rcvd_idx
        else:
            num_prefetched = 0
        self._saved_iter = len(self) - num_rest - num_prefetched

        return {'saved_iter': self._saved_iter,
                'sampler': self._sampler_to_save}

    def load_state_dict(self, state_dict):
        if state_dict is None:
            return
        self._need_to_resume = True
        self._saved_iter = state_dict['saved_iter']
        self._saved_sampler = state_dict['sampler']
        if dist.is_available() and type(self.sampler) is torch.utils.data.distributed.DistributedSampler:
            self._saved_sampler = [_[dist.get_rank()] for _ in self._saved_sampler]



def test1():
    train_loader = ResumableDataLoader(list(range(1000)), #testDataset(),
                                       batch_size=10,
                                       shuffle=True,
                                       num_workers=2)

    
    dataloader_iter = iter(train_loader)

    batches = []
    for i, data in enumerate(dataloader_iter):
        print(i, data)
        batches.append(data)
        if i == 3:
            print('Save the state_dict after the third iteration')
            state_dict = train_loader.state_dict()
        if i == 20:
            print('End up here')
            break

    train_loader.load_state_dict(state_dict)
    print('-----')
    print('Resume from the fourth iteration')
    for i, data in enumerate(train_loader):
        print(i+4, data)
        if i+4 >= len(batches):
            print(i+4)
            break
        assert (batches[i+4] == data).all()

    print('--------------')

def test2():

    # test 2: edge case, when dataiter has gone to the end. Make sure no error
    train_loader = ResumableDataLoader(list(range(10)), #testDataset(),
                                    batch_size=10,
                                    shuffle=True,
                                    num_workers=2)

    dataloader_iter = iter(train_loader)
    for i, data in enumerate(dataloader_iter):
        print(data)
    state_dict = train_loader.state_dict()
    train_loader.load_state_dict(state_dict)
    for i, data in enumerate(train_loader):
        print(data)

def main_worker(rank, ngpus, port):

    dist.init_process_group(
            world_size=ngpus, rank=rank,
            backend='nccl', init_method='tcp://127.0.0.1:%d' %port,
    )
    torch.cuda.set_device(rank)

    train_dataset = list(range(10000))
    train_sampler = torch.utils.data.distributed.DistributedSampler(train_dataset)

    train_loader = ResumableDataLoader(
        train_dataset, batch_size=10, shuffle=(train_sampler is None),
        num_workers=4, pin_memory=True, sampler=train_sampler)

    dataloader_iter = iter(train_loader)

    batches = []
    for i, data in enumerate(dataloader_iter):
        print(rank, i, data)
        batches.append(data)
        if i == 3:
            print('Save the state_dict after the third iteration')
            state_dict = train_loader.state_dict()
        if i == 20:
            print('End up here')
            break

    train_loader.load_state_dict(state_dict)
    print('-----')
    print('Resume from the fourth iteration')
    for i, data in enumerate(train_loader):
        print(rank, i+4, data)
        if i+4 >= len(batches):
            print(i+4)
            break
        assert (batches[i+4] == data).all()

def test3():
    # distributed:
    import torch.distributed as dist
    import torch.utils.data.distributed
    import torch.multiprocessing as mp
    
    port = 11000
    mp.spawn(main_worker, nprocs=2, args=(2, port))


if __name__ == '__main__':
    test3()








import torch
from torch.utils.data.distributed import DistributedSampler
import math
import warnings
import itertools


class DeterministicDistributedSampler(DistributedSampler):

    def __init__(self, *args, **kwargs):
        """
        we always start the generator with seed 0,
        and then to restart from certain iteration, we just
        manually drain the iterator. This can make dataloader
        properly resumed.

        Big note: we assume set_epoch takes in the iteration!!!

        Args:
            global_batch_size: We need global_batch_size because we
                we need to know how many interations each epoch so
                that we drain the iterator correctly.
                and we also manually apply drop_last.
            batch_size: local batch size. since we know world size,
                we can drive the global_batch_size
        """
        self.global_batch_size = kwargs.pop('global_batch_size', 0)
        self.batch_size = kwargs.pop('batch_size', 0)

        super().__init__(*args, **kwargs)

        warnings.warn(
            'You are using a customized distributed sampler. Make sure you are feeding '
            'iteration number to the set_epoch function; In addition, batch_size has to be fixed!'
        )

        # intialize batch_size and global_batch_size.
        if self.batch_size != 0:
            assert self.global_batch_size == 0, 'don\'t set both batch_size and global_batch_size'
            self.global_batch_size = self.batch_size * self.num_replicas
        elif self.global_batch_size != 0:
            assert self.batch_size == 0, 'don\'t set both batch_size and global_batch_size'
            self.batch_size = self.global_batch_size // self.num_replicas
        else:
            assert False, 'batch_size or global_batch_size should be specified.'

        # define the generator. There will be one generator form the beginning to the end.
        self.generator = torch.Generator()
        self.generator.manual_seed(self.seed)

        self._drained = False # if False, we will drain the iterator according to current self.epoch

    def __len__(self):
        if not self._drained:
            iterations_in_one_epoch = self.num_samples // self.batch_size
            done_iteration_in_this_epoch = self.epoch % iterations_in_one_epoch
            return (iterations_in_one_epoch - done_iteration_in_this_epoch) * self.batch_size
        else:
            return self.num_samples // self.batch_size * self.batch_size

    # Modified from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/samplers/distributed_sampler.py#L12
    def __iter__(self):
        start = self.rank
        # drain the sampler
        if not self._drained:
            _i = 0
            _iter = iter(self._indices())
            while _i < self.epoch * self.batch_size:
                try:
                    next(_iter)
                except:
                    _iter = iter(self._indices())
                    next(_iter)
            self._drained = True
            return _iter
        else:
            return iter(self._indices())
    
    def __iter__(self):
        start = self.rank
        indices = self._indices()
        # drain the sampler
        if not self._drained:
            done_epochs = (self.epoch * self.batch_size) // len(indices)
            for i in range(done_epochs):
                indices = self._indices()
            done_iteration_in_this_epoch = (self.epoch * self.batch_size) % len(indices)
            self._drained = True
            return iter(indices[done_iteration_in_this_epoch:])
        else:
            return iter(indices)

    def _indices(self):
        """
        This function almost copy from original pytorch implementation;
        first change: use one generator
        second change: directly apply drop_last here. Because we want to make sure
        """
        if self.shuffle:
            g = self.generator
            indices = torch.randperm(len(self.dataset), generator=g).tolist()
        else:
            indices = list(range(len(self.dataset)))

        if not self.drop_last:
            # add extra samples to make it evenly divisible
            padding_size = self.total_size - len(indices)
            if padding_size <= len(indices):
                indices += indices[:padding_size]
            else:
                indices += (indices * math.ceil(padding_size / len(indices)))[:padding_size]
        else:
            # remove tail of data to make it evenly divisible.
            indices = indices[:self.total_size]
        assert len(indices) == self.total_size

        # subsample
        indices = indices[self.rank:self.total_size:self.num_replicas]
        assert len(indices) == self.num_samples

        # manual drop_last
        indices = indices[:self.num_samples // self.batch_size * self.batch_size]

        return indices


class InfiniteDistributedSampler(DistributedSampler):

    def __init__(self, *args, **kwargs):
        """
        Args:
            global_batch_size: since infinite indices will wrap,
                so it is possible that same images in one batch.
                We apply drop_last here in the sampler.
            determistic: we always start the generator with seed 0,
                and then to restart from certain iteration, we just
                manually drain the iterator. This can make dataloader
                properly resumed.
        """
        self.global_batch_size = kwargs.pop('global_batch_size', 0)
        self.batch_size = kwargs.pop('batch_size', 0)

        self.deterministic = kwargs.pop('deterministic', False)

        super().__init__(*args, **kwargs)

        # intialize batch_size and global_batch_size.
        if self.batch_size != 0:
            assert self.global_batch_size == 0, 'don\'t set both batch_size and global_batch_size'
            self.global_batch_size = self.batch_size * self.num_replicas
        elif self.global_batch_size != 0:
            assert self.batch_size == 0, 'don\'t set both batch_size and global_batch_size'
            self.batch_size = self.global_batch_size // self.num_replicas
        else:
            assert not self.deterministic, 'You have to specify batch_size or global_batch_size if determinstic is True'

    # Modified from https://github.com/facebookresearch/detectron2/blob/master/detectron2/data/samplers/distributed_sampler.py#L12
    def __iter__(self):
        start = self.rank
        if self.deterministic:
            for _i, idx in enumerate(itertools.islice(self._infinite_indices(), start, None, self.num_replicas)):
                if _i >= self.epoch * self.batch_size:  # Make sure the epoch is actually iteration
                    yield idx
        else:
            yield from itertools.islice(self._infinite_indices(), start, None, self.num_replicas)

    def _infinite_indices(self):
        g = torch.Generator()
        if self.deterministic:
            g.manual_seed(0)
        else:
            g.manual_seed(self.epoch)
        while True:
            if self.shuffle:
                indices = torch.randperm(len(self.dataset), generator=g).tolist()
            else:
                indices = list(range(len(self.dataset)))


            # add extra samples to make it evenly divisible
            indices += indices[:(self.total_size - len(indices))]
            assert len(indices) == self.total_size

            if self.global_batch_size != 0:
                # do what drop_last do, make it devisible by global_batch_size
                indices = indices[:self.total_size // self.global_batch_size * self.global_batch_size]

            yield from indices


import os
import pylab

def save_current_fig(fn):
    save_path = os.path.join('/home-nfs/rluo/rluo/figures', fn)
    fig = pylab.gcf()
    if not os.path.exists(os.path.dirname(save_path)):
        os.makedirs(os.path.dirname(save_path))
    fig.savefig(save_path)



import types
import os
from rtutils.sampler import DeterministicDistributedSampler
import warnings

import pytorch_lightning as pl


def _get_distributed_sampler(self, dataloader, shuffle, mode):
    # modified from https://github.com/PyTorchLightning/pytorch-lightning/blob/HEAD/pytorch_lightning/trainer/data_loading.py?q=replace_ddp_sampler#L217
    from pytorch_lightning.utilities import _TORCH_GREATER_EQUAL_1_6
    from pytorch_lightning.overrides.distributed import UnrepeatedDistributedSampler
    from pytorch_lightning.trainer.states import RunningStage
    kwargs = self.distributed_sampler_kwargs
    kwargs["shuffle"] = shuffle and not self.overfit_batches
    if _TORCH_GREATER_EQUAL_1_6:
        kwargs.setdefault("seed", int(os.getenv("PL_GLOBAL_SEED", 0)))
    cls = UnrepeatedDistributedSampler if mode == RunningStage.PREDICTING else DeterministicDistributedSampler
    if cls == DeterministicDistributedSampler:
        kwargs["batch_size"] = dataloader.batch_size
    sampler = cls(dataloader.dataset, **kwargs)

    # rt note here: we set_epoch before, because we want the progress bart start from 0/{remaining batches}, instead of 0/{full batches}. However, we have now a better way to handle progressbar.
    # I just leave a note here to remind myself
    # # If we want to know if it is the end of an epoch, we cannot set_epoch here. because we rely on fixed number of batches each epoch.
    # # sampler.set_epoch(self.total_batch_idx) # we set_epoch here so that the progress bar will know the correct size. In fact pl will set_epoch later and override this. This is just a spoiler for the trainer to know how much batches left for this epoch.
    return sampler


def patch_pl_trainer_with_deterministic_sampler(trainer):
    """
    When pytorch lightning is replacing distributed sampler, replace with my dterministic sampler.
    """
    assert trainer.accelerator_connector.is_distributed, 'Your trainer is not distributed. Cannot replace.'
    assert trainer.accelerator_connector.replace_sampler_ddp, 'Make sure you set replace_sampler_ddp to be True'
    # TODO, we also need to make sure the dataloader is not set some DistributedSampler, otherwise the replace_sampler will not be called either.
    trainer._get_distributed_sampler = types.MethodType(_get_distributed_sampler, trainer)


class ProgressBarPatch(pl.callbacks.ProgressBar):
    def on_train_epoch_start(self, trainer, pl_module):
        super().on_train_epoch_start(trainer, pl_module)
        # This manually set the progress bar to the middle according to the total_batch_idx.
        # asssume fix num batches in each epoch.
        self.main_progress_bar.update(trainer.total_batch_idx % trainer.num_training_batches)


def patch_progressbar(trainer):
    """
    Make progressbar start from middle if resuming from middle-epoch checkpoint.
    """
    # This on_train_epoch_start will run after progress_bar on_train_epoch_start
    # this manually set the progress bar to the middle according to the total_batch_idx.
    # asssume fix num batches in each epoch.
    for callback in trainer.callbacks:
        if isinstance(callback, pl.callbacks.progress.ProgressBar):
            old_on_train_epoch_start = callback.on_train_epoch_start
            def on_train_epoch_start(self, trainer, pl_module):
                old_on_train_epoch_start(trainer, pl_module)
                self.main_progress_bar.update(trainer.total_batch_idx % trainer.num_training_batches)
            callback.on_train_epoch_start = types.MethodType(on_train_epoch_start, callback)


class KeyboardInterruptModelCheckpoint(pl.callbacks.ModelCheckpoint):
    def on_keyboard_interrupt(self, trainer, pl_module):
        # Save model when keyboard interrupt
        filepath = os.path.join(self.dirpath, self.CHECKPOINT_NAME_LAST+'.ckpt')
        self._save_model(trainer, filepath=filepath)


def patch_model_checkpoint(trainer):
    """
    Save last checkpoint when key interrupt occurs.
    """
    for callback in trainer.callbacks:
        if isinstance(callback, pl.callbacks.ModelCheckpoint):
            old_on_keyboard_interrupt = callback.on_keyboard_interrupt
            def on_keyboard_interrupt(self, trainer, pl_module):
                old_on_keyboard_interrupt(trainer, pl_module)
                KeyboardInterruptModelCheckpoint.on_keyboard_interrupt(self, trainer, pl_module)
            callback.on_keyboard_interrupt = types.MethodType(on_keyboard_interrupt, callback)


def patch_on_save_checkpoint_every(trainer, checkpoint_every):
    """
    Save last checkpoint every {checkpoint_every} iteration.
    """
    for callback in trainer.callbacks:
        if isinstance(callback, pl.callbacks.ModelCheckpoint):
            old_on_train_batch_end = callback.on_train_batch_end
            def on_train_batch_end(self, trainer, pl_module, outputs, batch, batch_idx, dataloader_idx):
                old_on_train_batch_end(trainer, pl_module, outputs, batch, batch_idx, dataloader_idx)
                if (trainer.global_step + 1) % checkpoint_every == 0:
                    KeyboardInterruptModelCheckpoint.on_keyboard_interrupt(self, trainer, pl_module)
            callback.on_train_batch_end = types.MethodType(on_train_batch_end, callback)


def patch_checkpoint_connector(trainer):
    # save and load total_batch_idx
    old_restore_training_state = trainer.checkpoint_connector.restore_training_state
    def restore_training_state(self, checkpoint, load_optimizer_states: bool = True):
        old_restore_training_state(checkpoint, load_optimizer_states)
        self.trainer.total_batch_idx = checkpoint.get('total_batch_idx', 0)
    old_dump_checkpoint = trainer.checkpoint_connector.dump_checkpoint
    def dump_checkpoint(self, weights_only: bool = False) -> dict:
        checkpoint = old_dump_checkpoint(weights_only)
        checkpoint['total_batch_idx'] = self.trainer.total_batch_idx
        if not self.trainer.total_batch_idx % self.trainer.num_training_batches == 0:
            # end in the middle of a epoch
            checkpoint['epoch'] -= 1 # so that when resuming, the current_epoch will be the same as when saving.
        return checkpoint
    trainer.checkpoint_connector.restore_training_state = types.MethodType(restore_training_state, trainer.checkpoint_connector)
    trainer.checkpoint_connector.dump_checkpoint = types.MethodType(dump_checkpoint, trainer.checkpoint_connector)


class SetEpochCallback(pl.callbacks.Callback):
    def on_train_epoch_start(self, trainer, pl_module):
        trainer.train_dataloader.sampler.set_epoch(trainer.total_batch_idx)


def patch_set_epoch(trainer):
    """
    Instead of set_epoch according to current_epoch, set_epoch with total_batch_idx.
    This is intended to work with my determnistic_distributed_sampler.
    """
    old_on_train_epoch_start = trainer.train_loop.on_train_epoch_start
    def on_train_epoch_start(self, epoch):
        old_on_train_epoch_start(epoch)
        self.trainer.train_dataloader.sampler.set_epoch(self.trainer.total_batch_idx)
    trainer.train_loop.on_train_epoch_start = types.MethodType(on_train_epoch_start, trainer.train_loop)


def patch_data_connector(trainer):
    """
    The the initial_batch_idx from the train_loader enumerator respect middle-epoch resuming.
    So that the check_val_fx can work properly.
    """
    # set enumerate initial batch_idx according to the loader size.
    # copy and rewrite the function
    # def get_profiled_train_dataloader(self, train_dataloader):
    #     # We feed batch_idx because the model may resume of middle-of-epoch checkpoint.
    #     # the length of train_dataloader has been modified by set_epoch at this point.
    #     from pytorch_lightning.trainer.supporters import prefetch_iterator
    #     start_batch_idx = self.trainer.num_training_batches - len(train_dataloader)
    #     profiled_dl = self.trainer.profiler.profile_iterable(
    #         enumerate(prefetch_iterator(train_dataloader), start_batch_idx), "get_train_batch"
    #     )
    #     return profiled_dl
    old_get_profiled_train_dataloader = trainer.data_connector.get_profiled_train_dataloader
    # reuse the old function
    def get_profiled_train_dataloader(self, train_dataloader):
        # We feed batch_idx because the model may resume of middle-of-epoch checkpoint.
        # the length of train_dataloader has been modified by set_epoch at this point.
        start_batch_idx = self.trainer.num_training_batches - len(train_dataloader)
        old_profiled_dl = old_get_profiled_train_dataloader(train_dataloader)
        def profiled_dl():
            # we discard the old batch_idx, and use the new one.
            for batch_idx, (_, batch) in enumerate(old_profiled_dl, start_batch_idx):
                yield batch_idx, batch
        return profiled_dl()
    trainer.data_connector.get_profiled_train_dataloader = types.MethodType(get_profiled_train_dataloader, trainer.data_connector)


def patch_everything(trainer):
    warnings.warn('Patch everything will replace trainer\'s (and its members\') functions with something else. In general it should be fine. But make sure you are confortable with this.')
    patch_model_checkpoint(trainer)
    patch_set_epoch(trainer)
    patch_progressbar(trainer)
    patch_pl_trainer_with_deterministic_sampler(trainer)
    patch_checkpoint_connector(trainer)
    patch_data_connector(trainer)


def patch_everything_safer(trainer):
    """
    Compare to patch_everything, we remove the patches that can be implemented by callbacks.
    This need to work with all_callbacks() and KeyboardInterruptModelCheckpoint to work properly.

    For examples:
    callbacks = [
        KeyboardInterruptModelCheckpoint(...),
        *rtutils.pl_patch.all_callbacks(),
    ]
    trainer = pl.Trainer(
        ...,
        callbacks = callbacks,
        ...
    )
    rtutils.pl_patch.patch_everything_safer(trainer)
    """
    warnings.warn('Patch everything will replace trainer\'s (and its members\') functions with something else. In general it should be fine. But make sure you are confortable with this.')
    warnings.warn('To fully function, you would also want to include all the callbacks in pl_path.all_backs and use KeyboardInterruptModelCheckpoint to create ModelCheckpoint Instance')
    patch_pl_trainer_with_deterministic_sampler(trainer)
    patch_checkpoint_connector(trainer)
    patch_data_connector(trainer)


def all_callbacks():
    return [
        ProgressBarPatch(),
        SetEpochCallback(),
    ]


import setuptools

#with open("README.md", "r") as fh:
#   long_description = fh.read()

setuptools.setup(
    name="rtutils", # Replace with your own username
    version="0.2.1",
    author="Ruotian Luo",
    author_email="rluo@ttic.edu",
    entry_points={
        'console_scripts': [
            'rtdrive=rtutils.drive:main',
            'gd=rtutils_cli.gdrive_wrapper:main'
        ]
    },
#    description="A small example package",
#    long_description=long_description,
#    long_description_content_type="text/markdown",
#   url="https://github.com/pypa/sampleproject",
    packages=setuptools.find_packages(),
#   classifiers=[
#       "Programming Language :: Python :: 3",
#       "License :: OSI Approved :: MIT License",
#       "Operating System :: OS Independent",
#   ],
#   python_requires='>=3.6',
)
